@@ -2,29 +2,20 @@ package Test::Aggregate;
 
 use warnings;
 use strict;
-
-use Test::Builder::Module;
-use Test::More;
 use Carp 'croak';
 
-use File::Find;
-
-use vars qw(@ISA @EXPORT $VERSION);
-@ISA    = qw(Test::Builder::Module);
-@EXPORT = @Test::More::EXPORT;
-
-BEGIN { $ENV{TEST_AGGREGATE} = 1 };
-
-END {   # for VMS
-    delete $ENV{TEST_AGGREGATE};
-}
-
+use Test::More;
+use Test::Aggregate::Base;
+use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION);
+@ISA    = qw(Test::Aggregate::Base);
+@EXPORT = (@Test::More::EXPORT, 'run_this_test_program');
 # controls whether or not we show individual test program pass/fail
 my %VERBOSE = (
     none     => 0,
     failures => 1,
     all      => 2,
 );
+my $BUILDER = Test::Builder->new;
 
 =head1 NAME
 
@@ -32,11 +23,11 @@ Test::Aggregate - Aggregate C<*.t> tests to make them run faster.
 
 =head1 VERSION
 
-Version 0.35
+Version 0.36
 
 =cut
 
-$VERSION = '0.35';
+$VERSION = '0.36';
 
 =head1 SYNOPSIS
 
@@ -62,6 +53,21 @@ named C<run_the_tests>.  This allows us to load perl only once and related
 modules only once.  If you have modules which are expensive to load, this can
 dramatically speed up a test suite.
 
+=head1 DEPRECATION
+
+For a whole variety of reasons, tests run in BEGIN/CHECK/INIT/INIT blocks are
+now deprecated.  They cause all sorts of test sequence headaches.  Plus, they
+break the up-coming nested TAP work.  You will have a problem if you use this
+common idiom:
+
+ BEGIN {
+     use_ok 'My::Module' or die;
+ }
+
+Instead, just C<use> the module and put the C<use_ok> tests in a t/load.t file
+or something similar and B<don't> aggregate it.  See the following for more
+information: L<http://use.perl.org/~Ovid/journal/38974>.
+
 =head1 USAGE
 
 Create a separate directory for your tests.  This should not be a subdirectory
@@ -254,163 +260,7 @@ the constructor.
 
 =cut
 
-sub _code_attributes {
-    qw/
-        setup
-        teardown
-        startup
-        shutdown
-    /;
-}
-
-sub new {
-    my ( $class, $arg_for ) = @_;
-
-    unless ( exists $arg_for->{dirs} || exists $arg_for->{tests} ) {
-        Test::More::BAIL_OUT("You must supply 'dirs' or 'tests'");
-    }
-    if ( exists $arg_for->{tests} && 'ARRAY' ne ref $arg_for->{tests} ) {
-        Test::More::BAIL_OUT(
-            "Argument for Test::Aggregate 'tests' key must be an array reference"
-        );
-    }
-        
-    $arg_for->{test_nowarnings} = 1 unless exists $arg_for->{test_nowarnings};
-    $arg_for->{set_filenames}   = 1 unless exists $arg_for->{set_filenames};
-    $arg_for->{findbin}         = 1 unless exists $arg_for->{findbin};
-    my $dirs = delete $arg_for->{dirs};
-    if ( defined $dirs ) {
-        $dirs = [$dirs] if 'ARRAY' ne ref $dirs;
-    }
-    else {
-        $dirs = [];
-    }
-
-    my $matching = qr//;
-    if ( $arg_for->{matching} ) {
-        $matching = delete $arg_for->{matching};
-        unless ( 'Regexp' eq ref $matching ) {
-            croak("Argument for 'matching' must be a pre-compiled regex");
-        }
-    }
-
-    my $has_code_attributes;
-    foreach my $attribute ( $class->_code_attributes ) {
-        if ( my $ref = $arg_for->{$attribute} ) {
-            if ( 'CODE' ne ref $ref ) {
-                croak("Attribute ($attribute) must be a code reference");
-            }
-            else {
-                $has_code_attributes++;
-            }
-        }
-    }
-
-    my $self = bless {
-        dirs            => $dirs,
-        matching        => $matching,
-        _no_streamer    => 0,
-        _packages       => [],
-    } => $class;
-    $self->{$_} = delete $arg_for->{$_} foreach (
-        qw/
-        check_plan
-        dry
-        dump
-        findbin
-        set_filenames
-        shuffle
-        test_nowarnings
-        tests
-        tidy
-        verbose
-        /,
-        $class->_code_attributes
-    );
-    $self->{tests} ||= [];
-
-    if ( my @keys = keys %$arg_for ) {
-        local $" = ', ';
-        croak("Unknown keys to &new:  (@keys)");
-    }
-
-    if ($has_code_attributes) {
-        eval "use Data::Dump::Streamer";
-        if ( my $error = $@ ) {
-            $self->{_no_streamer} = 1;
-            if ( my $dump = $self->_dump ) {
-                warn <<"                END_WARNING";
-Dump file ($dump) cannot be generated.  A code attributes was requested but
-we cannot load Data::Dump::Streamer:  $error.
-                END_WARNING
-                $self->{dump} = '';
-            }
-        }
-    }
-
-    return $self;
-}
-
-# set from user data
-
-sub _check_plan      { shift->{check_plan} || 0 }
-sub _dump            { shift->{dump} || '' }
-sub _dry             { shift->{dry} }
-sub _should_shuffle  { shift->{shuffle} }
-sub _matching        { shift->{matching} }
-sub _set_filenames   { shift->{set_filenames} }
-sub _findbin         { shift->{findbin} }
-sub _dirs            { @{ shift->{dirs} } }
-sub _startup         { shift->{startup} }
-sub _shutdown        { shift->{shutdown} }
-sub _setup           { shift->{setup} }
-sub _teardown        { shift->{teardown} }
-sub _tests           { @{ shift->{tests} } }
-sub _tidy            { shift->{tidy} }
-sub _test_nowarnings { shift->{test_nowarnings} }
-
-sub _verbose        {
-    my $self = shift;
-    $self->{verbose} ? $self->{verbose} : 0;
-}
-
-# set from internal data
-sub _no_streamer    { shift->{_no_streamer} }
-sub _packages       { @{ shift->{_packages} } }
-
-sub _get_tests {
-    my $self = shift;
-    my @tests;
-    my $matching = $self->_matching;
-    if ( $self->_dirs ) {
-        find( {
-                no_chdir => 1,
-                wanted   => sub {
-                    push @tests => $File::Find::name if /\.t\z/ && /$matching/;
-                }
-        }, $self->_dirs );
-    }
-    push @tests => $self->_tests;
-    
-    if ( $self->_should_shuffle ) {
-        $self->_shuffle(@tests);
-    }
-    else {
-        @tests = sort @tests;
-    }
-    return @tests;
-}
-
-sub _shuffle {
-    my $self = shift;
-
-    # Fisher-Yates shuffle
-    my $i = @_;
-    while ($i) {
-        my $j = rand $i--;
-        @_[ $i, $j ] = @_[ $j, $i ];
-    }
-    return;
+sub _do_dry_run {
 }
 
 sub run {
@@ -449,10 +299,12 @@ sub run {
     }
 
     $self->_startup->() if $self->_startup;
-    my $builder = Test::Builder->new;
 
-    # some tests may have been run in BEGIN blocks
-    $builder->{TEST_AGGREGATE_last_test} = @{ $builder->{Test_Results} } || 0;
+    # some tests may have been run in BEGIN blocks.  This is deprecated and
+    # now warns
+    my $tab = 'Test::Aggregate::Builder';
+    $BUILDER->{$tab}{last_test} = @{ $BUILDER->{Test_Results} } || 0;
+    $BUILDER->{$tab}{aggregate_program} = $self->{aggregate_program};
 
     my $current_test = 0;
     my @packages     = $self->_packages;
@@ -461,15 +313,15 @@ sub run {
         $current_test++;
         my ( $test, $package ) = @$data;
         $self->_setup->() if $self->_setup;
-        _run_this_test_program( $package => $test, $current_test, $total_tests, 2 );
+        run_this_test_program( $package => $test, $current_test, $total_tests, 2 );
         if ( my $error = $@ ) {
             Test::More::ok( 0, "Error running ($test):  $error" );
         }
 
         # XXX this should be fine since these keys are not actually used
         # internally.
-        $builder->{XXX_test_failed}       = 0;
-        $builder->{TEST_MOST_test_failed} = 0;
+        $BUILDER->{XXX_test_failed}       = 0;
+        $BUILDER->{TEST_MOST_test_failed} = 0;
         $self->_teardown->() if $self->_teardown;
     }
     $self->_shutdown->() if $self->_shutdown;
@@ -479,8 +331,11 @@ sub _any_tests_failed {
     my $failed  = 0; 
     my $builder = Test::Builder->new;
     my @summary = $builder->summary;
-    foreach my $passed (@summary[ $builder->{TEST_AGGREGATE_last_test} ..
-        $builder->current_test - 1 ]) {
+    foreach my $passed (
+        @summary[ $builder->{'Test::Aggregate::Builder'}{last_test} 
+            ..
+        $builder->current_test - 1 ]
+    ) {
         if (not $passed) {
             $failed = 1;
             last;
@@ -489,13 +344,13 @@ sub _any_tests_failed {
     return $failed;
 }
 
-sub _run_this_test_program {
+sub run_this_test_program {
     my $builder = Test::Builder->new;
     my ( $package, $test, $current_test, $num_tests, $verbose ) = @_;
     Test::More::diag("******** running tests for $test ********") if $ENV{TEST_VERBOSE};
     my $error = eval { 
-        if ( my $reason = $Test::Aggregate::Builder::SKIP_REASON_FOR{$package} ) {
-            Test::Builder->new->skip($reason);
+        if ( my $reason = $builder->{'Test::Aggregate::Builder'}{skip_reason_for}{$package} ) {
+            $builder->skip($reason);
             return;
         }
         else {
@@ -509,7 +364,7 @@ sub _run_this_test_program {
             local $|   = $|;
             local %SIG = %SIG;
             use warnings 'uninitialized';
-            $Test::Aggregate::Builder::FILE_FOR{$package} = $test;
+            $builder->{'Test::Aggregate::Builder'}{file_for}{$package} = $test;
             eval { $package->run_the_tests };
             $@;
         }
@@ -532,14 +387,15 @@ sub _run_this_test_program {
             $builder->{TEST_MOST_test_failed} = 0;
         }
     }
-    $builder->{TEST_AGGREGATE_last_test} = $builder->current_test;
+    $builder->{'Test::Aggregate::Builder'}{last_test} = $builder->current_test;
 
     return unless $error;
 }
 
 sub _build_aggregate_code {
     my ( $self, @tests ) = @_;
-    my $code = $self->_test_builder_override;
+    my $code = "\n# Built from $0\n";
+    $code .= $self->_test_builder_override;
 
     my ( $startup,  $startup_code )  = $self->_as_code('startup');
     my ( $shutdown, $shutdown_code ) = $self->_as_code('shutdown');
@@ -593,14 +449,15 @@ $findbin
         #$test_code =~ s/\n__(?:DATA|END)__\n$//s;
 
         if ( $test_code =~ /^(__(?:DATA|END)__)/m ) {
-            Test::More::BAIL_OUT("Test $test not allowed to have $1 token");
+            Test::More::BAIL_OUT("Test $test not allowed to have $1 token (Test::Aggregate::Nested supports them)");
         }
+
         my $package   = $self->_get_package($test);
         push @{ $self->{_packages} } => [ $test, $package ];
         if ( $setup ) {
             $code .= "    $setup->('$test');\n";
         }
-        $code .= qq{    Test::Aggregate::_run_this_test_program( $package => "$test", $current_test, $total_tests, $verbose );};
+        $code .= qq{    run_this_test_program( $package => "$test", $current_test, $total_tests, $verbose );};
 
         if ( $teardown ) {
             $code .= "    $teardown->('$test');\n";
@@ -610,6 +467,17 @@ $findbin
         my $set_filenames = $self->_set_filenames
             ? "local \$0 = '$test';"
             : '';
+
+        if ( $self->_check_plan ) {
+            $test_code .= <<"            END";
+my \$tab   = Test::Builder->new->{'Test::Aggregate::Builder'};
+my \$plan  = \$tab->{plan_for}{'$package'} || 0;
+my \$file  = \$tab->{file_for}{'$package'};
+my \$tests = \$tab->{tests_run}{'$package'} || 0;
+Test::More::is( \$tests, \$plan, "Test (\$file) should have the correct plan" );
+            END
+        }
+
         $test_packages .= <<"        END_CODE";
 {
 $separator beginning of $test $separator
@@ -627,8 +495,9 @@ $separator end of $test $separator
     if ( $shutdown ) {
         $code .= "    $shutdown->() if __FILE__ eq '$dump';\n";
     }
-
-    $code .= "}\n$test_packages";
+    
+    $code .= "}\n";
+    $code .= $test_packages;
     if ( my $tidy = $self->_tidy ) {
         eval "use Perl::Tidy";
         my $error = $@;
@@ -675,12 +544,6 @@ sub _slurp {
     return do { local $/; <FH> };
 }
 
-sub _get_package {
-    my ( $class, $file ) = @_;
-    $file =~ s/\W//g;
-    return $file;
-}
-
 sub _test_builder_override {
     my $self = shift;
 
@@ -695,10 +558,11 @@ BEGIN {
     *Test::NoWarnings::had_no_warnings = sub { };
     *Test::NoWarnings::import = sub {
         my $callpack = caller();
-        if ( $Test::Aggregate::Builder::PLAN_FOR{$callpack} ) {
-            $Test::Aggregate::Builder::PLAN_FOR{$callpack}--;
+        my $ta_builder = $BUILDER->{'Test::Aggregate::Builder'};
+        if ( $ta_builder->{plan_for}{$callpack} ) {
+            $ta_builder->{plan_for}{$callpack}--;
         }
-        $Test::Aggregate::Builder::TEST_NOWARNINGS_LOADED{$callpack} = 1;
+        $ta_builder->{test_nowarnings_loaded}{$callpack} = 1;
     };
 }
         END_CODE
@@ -707,10 +571,12 @@ BEGIN {
     return <<"    END_CODE";
 use Test::Aggregate;
 use Test::Aggregate::Builder;
-BEGIN { \$Test::Aggregate::Builder::CHECK_PLAN = $check_plan };
+my \$BUILDER;
+BEGIN { 
+    \$BUILDER = Test::Builder->new;
+    \$BUILDER->{'Test::Aggregate::Builder'}{check_plan} = $check_plan 
+};
 $disable_test_nowarnings;
-
-my \$BUILDER = Test::Builder->new;
     END_CODE
 }
 
@@ -828,9 +694,16 @@ tests, there is no need to run them with this.  Otherwise:
 
 =over 4
 
+=item * C<exit>
+
+Don't call exit() in your aggregated tests.  We now warn very verbosely if
+this is done, but we still exit on the assumption that further tests cannot
+run.
+
 =item * C<__END__> and C<__DATA__> tokens.
 
 These won't work and the tests will call BAIL_OUT() if these tokens are seen.
+However, this limitation does not apply to L<Test::Aggregate::Nested>.
 
 =item * C<BEGIN> and C<END> blocks.
 
@@ -977,7 +850,7 @@ out which one is actually causing the failure.
 
 =head1 COMMON PITFALLS
 
-=head2 My Tests Through an Exception But Passed Anyway!
+=head2 My Tests Threw an Exception But Passed Anyway!
 
 This really isn't a C<Test::Aggregate> problem so much as a general Perl
 problem.  For each test file, C<Test::Aggregate> wraps the tests in an eval
